% Text Mining with MATLAB 2nd Edition
% www.http://www.textmininglab.net/

% CHAPTER 14: Keyword Extraction and Summarization

disp('Example 14.1 -----------------------------------------------------');
clear                                                              %(14.1a)
load datasets_ch14a
whos                                                               %(14.1b)
disp(data.Properties.VariableNames)                                %(14.1c)
pause;

disp('Example 14.2 -----------------------------------------------------');
% creates an array of tokenized documents                           %(14.2)
tokdocs = tokenizedDocument(data.text);
vocab = tokdocs.Vocabulary; % extracts the vocabulary
% identifies short words (with less than 5 characters)
shortwords = vocab(strlength(vocab)<5);
% removes short words and stop words
prepdocs = removeWords(tokdocs,union(shortwords,stopWords));
pause;

disp('Example 14.3 -----------------------------------------------------');
hf = figure(1);                                                     %(14.3)
set(hf,'Color',[1 1 1],'Name','Word Cloud of frequent words'); 
wordcloud(prepdocs,'Shape','oval','HighlightColor','blue',...      
          'MaxDisplayWords',150,'SizePower',0.2);
pause;

disp('Example 14.4 -----------------------------------------------------');
sentence = "This illustrates a keyword extraction example "+...     %(14.4)
           "with the RAKE algorithm for keyword extraction.";
keywords = rakeKeywords(tokenizedDocument(sentence))
pause;

disp('Example 14.5 -----------------------------------------------------');
keywords.Keyword = strip(join(keywords.Keyword))                    %(14.5)
pause

disp('Example 14.6 -----------------------------------------------------');
keywords = rakeKeywords(tokdocs);                                   %(14.6)
keywords.Keyword = strip(join(keywords.Keyword));
height(keywords) % number of keywords extracted
pause;

disp('Example 14.7 -----------------------------------------------------');
% Keywords extracted from first sentence                            %(14.7)
disp(keywords(keywords.DocumentNumber==1,:))
pause;

disp('Example 14.8 -----------------------------------------------------');
[uniquekeywords,~,uidx] = unique(keywords.Keyword);                 %(14.8)
keywordfreqs = hist(uidx,length(uniquekeywords));
[rankedfreqs,ridx] = sort(keywordfreqs,'descend');
rankedkeywords = uniquekeywords(ridx);
pause;

disp('Example 14.9 -----------------------------------------------------');
hf = figure(3);                                                     %(14.9)
set(hf,'Color',[1 1 1],'Name','Word Cloud of RAKE Keywords'); 
wordcloud(rankedkeywords,rankedfreqs,'HighlightColor','red',...
          'MaxDisplayWords',150,'SizePower',0.2);
pause;

disp('Example 14.10 ----------------------------------------------------');
sentence = "This illustrates a keyword extraction example "+...    %(14.10)
           "with the TextRank algorithm for keyword extraction.";
keywords = textrankKeywords(tokenizedDocument(sentence),'Window',3)
pause;

disp('Example 14.11 ----------------------------------------------------');
disp('not a MATLAB script')
pause;

disp('Example 14.12 ----------------------------------------------------');
kwds ={'keyword','extraction','example','algorithm','TextRank'};   %(14.12)
nodes1 = [1 1 1 2 4];
nodes2 = [2 3 4 3 5];
pause;

disp('Example 14.13 ----------------------------------------------------');
kwdsgraph = graph(nodes1,nodes2,[],kwds);                          %(14.13)
pageranks = centrality(kwdsgraph,'pagerank')'
pause;

disp('Example 14.14 ----------------------------------------------------');
textranks = pageranks * length(kwds)                               %(14.14)
pause;

disp('Example 14.15 ----------------------------------------------------');
% identifies words with entity labels different from 'person'     %(14.15a)
newtokdocs = addEntityDetails(tokdocs); 
details = tokenDetails(newtokdocs);
nonpersons = details.Token(details.Entity~='person');
% creates a bag of words model                                    %(14.15b)
bow = bagOfWords(newtokdocs);
bow = removeWords(bow,nonpersons); % removes non-persons
bow = removeInfrequentWords(bow,5); % removes infrequent persons
% computes the co-occurrence matrix and builds the graph          %(14.15c)
counts = bow.Counts;
cooccurrencemtx = counts.'*counts;
pgraph = graph(cooccurrencemtx,bow.Vocabulary,'omitselfloops');
hf = figure(5); % plots the graph                                 %(14.15d)
figname = 'Graph of most frequent characters in Don Quixote';
set(hf,'Color',[1 1 1],'Name',figname);
plot(pgraph); axis off
pause;

disp('Example 14.16 ----------------------------------------------------');
pgraph.Nodes.Degree = centrality(pgraph,'degree');                 %(14.16)
pgraph.Nodes.Closeness = centrality(pgraph,'closeness');
pgraph.Nodes.Betweenness = centrality(pgraph,'betweenness');
pgraph.Nodes.PageRank = centrality(pgraph,'pagerank');
disp(pgraph.Nodes)
pause;

disp('Example 14.17 ----------------------------------------------------');
ch1sentences = tokenizedDocument(data(data.chap==1,:).text);       %(14.17)
lexrnkscores = lexrankScores(ch1sentences);
txtrnkscores = textrankScores(ch1sentences);
pause;

%%%% GENERATES FIGURE 14.6
hf = figure(6);                                
figname = 'LexRank and TextRank scores for sentences in chapter 1';
set(hf,'Color',[1 1 1],'Name',figname);
bar(1:length(ch1sentences),[lexrnkscores,txtrnkscores])
legend({'LexRank Scores','TextRank Scores'},'Location','northwest')
pause
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


disp('Example 14.18 ----------------------------------------------------');
[~,idxlexrnk] = sort(lexrnkscores,'descend');                      %(14.18)
[~,idxtxtrnk] = sort(txtrnkscores,'descend');
disp(length(intersect(idxlexrnk(1:10),idxtxtrnk(1:10))))
pause;

disp('Example 14.19 ----------------------------------------------------');
disp('not a MATLAB script')
pause;

disp('Example 14.20 ----------------------------------------------------');
load datasets_ch14b                                               %(14.20a)
summaries.Properties.VariableNames
height(summaries)                                                 %(14.20b)
pause;

% TABLE GENERATION
for k = 1:size(summaries,1)
    nsent(k) = length(splitSentences(summaries.text(k)));
    span(k) = sprintf("%d to %d",summaries.from(k),summaries.to(k));
end
tbl = table();
tbl.episode = summaries.episode;
tbl.span = span';
tbl.summary_size = nsent';
tbl
pause
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Example 14.21 ----------------------------------------------------');
reference = summaries.text(1);                                     %(14.21)
referencetok = tokenizedDocument(reference);
pause;

disp('Example 14.22 ----------------------------------------------------');
chapterspan = (data.chap>=summaries.from(1))&...                   %(14.22)
              (data.chap<=summaries.to(1));
episodetok = tokdocs(chapterspan);
pause;

disp('Example 14.23 ----------------------------------------------------');
referencesize = length(splitSentences(reference));                %(14.23a)
% extracts summary with LexRank scores                            %(14.23b)
extractedlex = extractSummary(episodetok,'ScoringMethod','lexrank',...
                              'SummarySize',referencesize);
summarylextok = tokenizedDocument(join(joinWords(extractedlex)));
% extracts summary with TextRank scores                           %(14.23c)
extractedtxt = extractSummary(episodetok,'ScoringMethod','textrank',...
                              'SummarySize',referencesize);
summarytxttok = tokenizedDocument(join(joinWords(extractedtxt)));
pause;

disp('Example 14.24 ----------------------------------------------------');
rougelex = rougeEvaluationScore(summarylextok,referencetok)       %(14.24a)
rougetxt = rougeEvaluationScore(summarytxttok,referencetok)       %(14.24b)
pause;

disp('Example 14.25 ----------------------------------------------------');
rng('default');                                                   %(14.25a)
randsummscore =[];
for k=1:500                                                       %(14.25b)
    % generates a random summary
    randindx = randperm(length(episodetok));
    rndsumm = episodetok(sort(randindx(1:referencesize)));
    rndsumm = tokenizedDocument(join(joinWords(rndsumm)));
    % computes its ROUGE score
    randsummscore(k) = rougeEvaluationScore(rndsumm,referencetok);
end
hf = figure(7);                                                   %(14.25c)
figname = 'Distribution of ROUGE scores for random summaries';
set(hf,'Color',[1 1 1],'Name',figname);
histogram(randsummscore,'Normalization','probability');
text(0.42,0.12,sprintf('mean = %6.4f',mean(randsummscore)));
text(0.42,0.11,sprintf('std = %6.4f',std(randsummscore)));
pause;

disp('Example 14.26 ----------------------------------------------------');
summarysizes = 1:30;                                              %(14.26a)
meanvalues = [];
stdvalues = [];
txtrnkscore = [];
for k=1:length(summarysizes)                                      %(14.26b)
    randsummscore = [];
    for n=1:100
        % generates and evaluates random summaries
        randindx = randperm(length(episodetok));
        rndsumm = episodetok(sort(randindx(1:summarysizes(k))));
        rndsumm = tokenizedDocument(join(joinWords(rndsumm)));
        randsummscore(n) = rougeEvaluationScore(rndsumm,referencetok);
    end
    meanvalues(k) = mean(randsummscore);
    stdvalues(k) = std(randsummscore);
    % generates and evaluates TextRank summaries 
    extracted = extractSummary(episodetok,'SummarySize',summarysizes(k));
    summarytok = tokenizedDocument(join(joinWords(extracted)));
    txtrnkscore(k) = rougeEvaluationScore(summarytok,referencetok);
end
pause;

%%%%%%%%%%% GENERATES THE PLOT%%%%%%%%%%%%%
hf = figure(8);
figname = 'ROUGE scores for summaries of different lengths';
set(hf,'Color',[1 1 1],'Name',figname);
plot(summarysizes,txtrnkscore,'.-.k');
hold on
plot([1,30],[rougetxt,rougetxt],'-.k');
plot(summarysizes,meanvalues,'.-b');
plot(summarysizes,meanvalues+stdvalues,':b');
plot(summarysizes,meanvalues-stdvalues,':b');
hold off
ylabel('ROUGE Score')
xlabel('Length of Summaries (in number of sentences)')
legendlabels = {'ROUGE scores for TextRank summaries',...
                'ROUGE score for TextRank summary of 12 sentences',...
                'Mean values of ROUGE for random summaries',...
                'Mean values of ROUGE plus/minus standard deviations'};
legend(legendlabels,'Location','southeast')
pause;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%% Generate Figure 14.9 %%%%%%%%%
fsscore=[]; rsscore=[]; fs2score=[];
for k=1:size(summaries,1), fprintf('.');
    summary = summaries.text(k);
    nsent = length(splitSentences(summary));
    ps = tokenizedDocument(summary);
    
    ini = summaries.from(k);
    fin = summaries.to(k);
    span = (data.chap>=ini)&(data.chap<=fin);
    tokdocs = tokenizedDocument(data(span,:).text);
    [summ,scores] = extractSummary(tokdocs,...
    'SummarySize',nsent,'OrderBy','position');
    fs = tokenizedDocument(join(joinWords(summ)));
    fsscore(k) = rougeEvaluationScore(fs,ps);
    
    [summ,scores] = extractSummary(tokdocs,...
    'ScoringMethod','lexrank','SummarySize',nsent,'OrderBy','position');
    fs2 = tokenizedDocument(join(joinWords(summ)));
    fs2score(k) = rougeEvaluationScore(fs2,ps);
    
    temp=[];
    for n=1:100
        randindx = randperm(length(tokdocs));
        rs = tokdocs(sort(randindx(1:nsent)));
        rs = tokenizedDocument(join(joinWords(rs)));
    	temp(n) =rougeEvaluationScore(rs,ps);
    end
    rsscore(k) = mean(temp);
end, fprintf('\n');

hf = figure(9)
figname = 'ROUGE scores for summaries of episodes in Table 14.1';
set(hf,'Color',[1 1 1],'Name',figname);
bar([1:9],[fsscore',fs2score',rsscore'])
legend({'TextRank','LexRank','Random (average)'},'Orientation','horizontal')
ax = axis; axis([ax(1:3),0.8])
ylabel('ROUGE Score')

[mean(fsscore),mean(fs2score),mean(rsscore)]
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
